SystemOrganization addCategory: #Unification!
SystemOrganization addCategory: #'Unification-Destructuring'!
SystemOrganization addCategory: #'Unification-Tests-Destructuring'!
SystemOrganization addCategory: #'Unification-Tests'!

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 1/12/2012 09:23'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ self children
		inject: anMGU
		into: [:mgu :each | mgu addAll: (solutionFinder findSolution: each using: mgu)]! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:31'!
isMetaVariable
	^ self unapply first = #Variable.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 8/9/2011 07:32'!
or: anObject
	^ OrUnifier left: self right: anObject! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 1/4/2012 20:20'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].

	^ anObject unificationClosureWithAlgebraic: self in: termRelation.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 1/11/2012 20:16'!
unificationClosureWithAlgebraic: anObject in: termRelation
	"s = f(s1, ..., sm), t = g(t1, ..., tn) where m, n > 0"
	| partition |
	(self type = anObject type) ifFalse: [^ self failToUnifyWith: anObject. "f ~= g"].
	(self arity = anObject arity) ifFalse: [^ self failToUnifyWith: anObject. "m ~= n"].
	partition := termRelation.
	self children with: anObject children do:
		[:si :ti | partition := (partition find: si) unificationClosureWith: (partition find: ti) in: partition].
	^ partition.! !

AlgebraicDataType subclass: #VariadicTerm
	instanceVariableNames: 'subterms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!VariadicTerm commentStamp: 'fbs 5/29/2011 16:47' prior: 0!
A general term that may contain any (variable) number of subterms. It's mainly just for testing.!

!VariadicTerm class methodsFor: 'instance variables' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	^ self new with: aCollection.! !

!VariadicTerm methodsFor: 'destructuring' stamp: 'fbs 5/27/2011 20:27'!
unapply
	^ #Term asOrderedCollection, subterms.! !

!VariadicTerm methodsFor: 'private' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	subterms := aCollection! !

Object subclass: #DestructuringUnifier
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Destructuring'!

!DestructuringUnifier commentStamp: 'fbs 1/10/2012 09:25' prior: 0!
A DestructuringUnifier is an object used to "destructure" an object, that is, to tear some object apart in some way. For instance, the classic Prolog statement [H | T] = [1, 2, 3] might be translated as (ConcatenatedList headNamed: #h tailNamed: #t) =? #(1 2 3).!

DestructuringUnifier subclass: #HeadUnifier
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Destructuring'!

!HeadUnifier commentStamp: 'fbs 10/13/2011 07:29' prior: 0!
I represent the head of an OrderedCollection.!

!HeadUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:30'!
named: aSymbol
	^ self value: aSymbol asVariable.! !

!HeadUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:31'!
value: anObject
	^ self new value: anObject.! !

!HeadUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/11/2012 14:01'!
values: someObjects
	^ MultiHeadUnifier new values: someObjects.! !

!HeadUnifier methodsFor: 'unification' stamp: 'fbs 1/12/2012 09:34'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ solutionFinder findSolution: value using: anMGU.! !

!HeadUnifier methodsFor: 'unification' stamp: 'fbs 1/14/2012 21:06'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].
	anObject isCollection
		ifFalse: [^ self failToUnifyWith: anObject].
	^ self unificationClosureWithSequence: anObject in: termRelation! !

!HeadUnifier methodsFor: 'unification' stamp: 'fbs 1/14/2012 21:18'!
unificationClosureWithSequence: aSequence in: termRelation
	aSequence ifEmpty: [^ self failToUnifyWith: aSequence].
	^ value unificationClosureWith: aSequence first in: termRelation.! !

!HeadUnifier methodsFor: 'evaluating' stamp: 'fbs 1/14/2012 21:13'!
value
	^ value.! !

!HeadUnifier methodsFor: 'private' stamp: 'fbs 10/13/2011 07:31'!
value: anObject
	value := anObject.! !

!HeadUnifier methodsFor: 'structuring' stamp: 'fbs 1/16/2012 13:13'!
| aTail
	^ ListUnifier head: self value tail: aTail.! !

HeadUnifier subclass: #MultiHeadUnifier
	instanceVariableNames: 'values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Destructuring'!

MultiHeadUnifier class
	instanceVariableNames: 'values'!

!MultiHeadUnifier commentStamp: 'fbs 1/11/2012 18:51' prior: 0!
Just like a HeadUnifier, but I unify against the first N elements of some collection.!

MultiHeadUnifier class
	instanceVariableNames: 'values'!

!MultiHeadUnifier methodsFor: 'private' stamp: 'fbs 1/14/2012 21:18'!
unificationClosureWithSequence: aSequence in: termRelation
	aSequence ifEmpty: [^ self failToUnifyWith: aSequence].
	^ values unificationClosureWith: (aSequence first: values size) in: termRelation.! !

!MultiHeadUnifier methodsFor: 'private' stamp: 'fbs 1/11/2012 18:43'!
values: someObjects
	values := someObjects.! !

DestructuringUnifier subclass: #ListUnifier
	instanceVariableNames: 'head tail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Destructuring'!

!ListUnifier commentStamp: 'fbs 1/11/2012 18:53' prior: 0!
I destructure a (sequenceable) collection into a head and a tail. For instance,

    ListUnifier headNamed: #h tailNamed: #t =? #(1 2 3)

unifies h with 1 and t with #(2 3).!

!ListUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:35'!
head: anObject tail: anotherObject
	^ self new head: anObject tail: anotherObject.! !

!ListUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/11/2012 19:02'!
head: anObject tailNamed: tailSymbol
	^ self new head: anObject tail: tailSymbol asVariable.! !

!ListUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/11/2012 19:01'!
headNamed: headSymbol tail: anObject
	^ self new head: headSymbol asVariable tail: anObject.! !

!ListUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:36'!
headNamed: headSymbol tailNamed: tailSymbol
	^ self new head: headSymbol asVariable tail: tailSymbol asVariable.! !

!ListUnifier methodsFor: 'unification' stamp: 'fbs 1/12/2012 09:39'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ {head. tail.}
		inject: anMGU
		into: [:mgu :each | solutionFinder findSolution: each using: mgu].
! !

!ListUnifier methodsFor: 'accessing' stamp: 'fbs 1/14/2012 20:36'!
head
	^ head.! !

!ListUnifier methodsFor: 'private' stamp: 'fbs 10/13/2011 07:35'!
head: anObject tail: anotherObject
	head := anObject.
	tail := anotherObject.! !

!ListUnifier methodsFor: 'accessing' stamp: 'fbs 1/14/2012 20:36'!
tail
	^ tail.! !

!ListUnifier methodsFor: 'unification' stamp: 'fbs 1/14/2012 15:27'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].
	anObject isCollection
		ifFalse: [^ self failToUnifyWith: anObject].
	anObject isEmpty
		ifTrue: [^ self failToUnifyWith: anObject].
		
	^ self unificationClosureWithSequence: anObject in: termRelation! !

!ListUnifier methodsFor: 'unification' stamp: 'fbs 1/14/2012 15:27'!
unificationClosureWithSequence: someObjects in: termRelation
	| h t partition |
	"It would be nice to drive the type checking into head and tail. The trouble is that tail needs to know head's size. Rather than polluting tail's API, we hold our nose and suffer the below."
	"Calculating headSize = 1 if not a collection doesn't help because if size = 1, there's still a difference between x and {x}."
	h := head isCollection
		ifTrue: [someObjects first: head size]
		ifFalse: [1].
	t := head isCollection
		ifTrue: [someObjects allButFirst: head size]
		ifFalse: [someObjects allButFirst].
	partition := head unificationClosureWith: h in: termRelation.
	^ tail unificationClosureWith: t in: partition.! !

DestructuringUnifier subclass: #TailUnifier
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Destructuring'!

!TailUnifier commentStamp: 'fbs 10/13/2011 07:33' prior: 0!
I represent the tail (or rest) of an OrderedCollection.!

!TailUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:33'!
named: aSymbol
	^ self value: aSymbol asVariable.! !

!TailUnifier class methodsFor: 'instance creation' stamp: 'fbs 10/13/2011 07:33'!
value: anObject
	^ self new value: anObject.! !

!TailUnifier methodsFor: 'unification' stamp: 'fbs 1/12/2012 09:40'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ solutionFinder findSolution: value using: anMGU.! !

!TailUnifier methodsFor: 'unification' stamp: 'fbs 1/11/2012 19:20'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].
	anObject isCollection
		ifFalse: [^ self failToUnifyWith: anObject].
	anObject isEmpty
		ifTrue: [^ self failToUnifyWith: anObject].
	^ value unificationClosureWith: anObject allButFirst in: termRelation.! !

!TailUnifier methodsFor: 'unification' stamp: 'fbs 1/11/2012 19:20'!
unificationClosureWithSequence: aSequence in: termRelation
	aSequence ifEmpty: [^ self failToUnifyWith: aSequence].
	^ value unificationClosureWith: aSequence allButFirst in: termRelation.! !

!TailUnifier methodsFor: 'evaluating' stamp: 'fbs 1/14/2012 21:19'!
value
	^ value.! !

!TailUnifier methodsFor: 'private' stamp: 'fbs 10/13/2011 07:33'!
value: anObject
	value := anObject.! !

Object subclass: #MetaVariable
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!MetaVariable commentStamp: 'fbs 5/24/2011 11:04' prior: 0!
A MetaVariable is a reified variable, that is, an object representing a variable. It is a placeholder in a structure that says "we don't know what the value is of this part of the structure, but we can give it a name."!

!MetaVariable class methodsFor: 'instance creation' stamp: 'fbs 5/24/2011 07:31'!
named: aSymbol
	^ self new name: aSymbol.! !

!MetaVariable methodsFor: 'comparing' stamp: 'fbs 1/4/2012 20:33'!
= anObject
	^ anObject isMetaVariable and: [name = anObject name]! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 22:06'!
children
	^ #().! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:05'!
hasChildren
	"Even though unapplied a variable looks like it has a parameter, that's just its name."
	^ false.! !

!MetaVariable methodsFor: 'comparing' stamp: 'fbs 1/4/2012 20:34'!
hash
	"17's a prime."
	"Implementation copied from AlgebraicDataType >> #hash."
	^ self unapply inject: 17 into: [:acc :each | 17 * each hash].! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 1/4/2012 20:16'!
isMetaVariable
	^ true.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 07:30'!
name
	^ name.! !

!MetaVariable methodsFor: 'private' stamp: 'fbs 5/24/2011 07:32'!
name: aSymbol
	name := aSymbol.! !

!MetaVariable methodsFor: 'printing' stamp: 'fbs 1/4/2012 20:17'!
printOn: aStream
	aStream nextPut: $(.
	aStream nextPutAll: ((self unapply collect: [:token | token printString])
		reduce: [:acc :each | acc , ' ', each]).
	aStream nextPut: $).! !

!MetaVariable methodsFor: 'destructuring' stamp: 'fbs 5/24/2011 07:34'!
unapply
	^ {#Variable. name.}! !

!MetaVariable methodsFor: 'unification' stamp: 'fbs 1/3/2012 15:30'!
unificationClosureWith: anObject in: termRelation
	^ termRelation union: self with: anObject.! !

Object subclass: #MostGeneralUnifier
	instanceVariableNames: 'dict'
	classVariableNames: 'EmptyMGU'
	poolDictionaries: ''
	category: 'Unification'!

!MostGeneralUnifier commentStamp: 'fbs 5/2/2012 09:49' prior: 0!
Given two structures s and t, each containing zero or more MetaVariables, to what values must those variables be set for s and t to be structurally equivalent?

If s := 1 and t := #x asVariable, then the most general unifier (mgu) will be {#x asVariable -> 1}.

If s := 2 and t := 1 then the mgu does not exist, and unification will raise a UnificationFailure.

If s := t := 1 then the mgu is {}, or "there's no unification needed".!

!MostGeneralUnifier class methodsFor: 'instance creation' stamp: 'fbs 5/2/2012 16:46'!
empty
	"A well-known object representing the empty most general unifier."
	^ EmptyMGU.! !

!MostGeneralUnifier class methodsFor: 'class initialization' stamp: 'fbs 5/2/2012 16:46'!
initialize
	EmptyMGU := self new.! !

!MostGeneralUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/12/2012 18:47'!
new
	^ self basicNew wrappingDictionary: Dictionary new.! !

!MostGeneralUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/12/2012 19:15'!
newFrom: someAssociations
	^ self new wrappingDictionary: (Dictionary newFrom: someAssociations).! !

!MostGeneralUnifier class methodsFor: 'instance creation' stamp: 'fbs 1/12/2012 18:43'!
newFromPairs: anArray 
	"Answer an instance of me associating (anArray at:i) to (anArray at: i+1)
	 for each odd i.  anArray must have an even number of entries."

	^ self new wrappingDictionary: (Dictionary newFromPairs: anArray).! !

!MostGeneralUnifier methodsFor: 'comparing' stamp: 'fbs 1/12/2012 19:16'!
= anMGU
	(anMGU isKindOf: self class) ifFalse: [^ false].
	^ self asArray = anMGU asArray.! !

!MostGeneralUnifier methodsFor: 'adding' stamp: 'fbs 5/2/2012 10:07'!
addAll: anMGU <MostGeneralUnifier>
	"Add all key->values in anMGU and self to a new MGU. When x->y and y->k (x, y MetaVariables) add x->k and y->k instead."
	| newMapping twoStepVariables |
	newMapping := Dictionary newFromPairs: self asArray.
	anMGU keys do: [:k |
		newMapping
			at: k
			ifPresent: [:v | MostGeneralUnifierMergeFailure  signal: ('{1} is already bound to {2}' format: {k. v})]
			ifAbsent: [newMapping at: k put: (anMGU at: k)]].
	
	twoStepVariables := (self values intersection: anMGU keys).
	"'Flatten' two-step mappings from x->y and y->k to x->k and y->k"
	twoStepVariables do: [:y |
		(self keysMappingTo: y) do: [:x | newMapping at: x put: (anMGU at: y)]].
	^ self class newFromPairs: (self asArray: newMapping).! !

!MostGeneralUnifier methodsFor: 'converting' stamp: 'fbs 2/10/2012 20:32'!
asArray
	^ self asArray: dict.! !

!MostGeneralUnifier methodsFor: 'converting' stamp: 'fbs 5/2/2012 09:52'!
asArray: aDictionary
	"Answer an Array whose elements are the elements of the receiver.
	Implementation note: Cannot use ''Array withAll: self'' as that only
	works for SequenceableCollections that support the replacement 
	primitive."

	| array index |
	array := Array new: (aDictionary size * 2).
	index := 0.
	aDictionary associationsDo: [:each |
		array at: (index := index + 1) put: each key.
		array at: (index := index + 1) put: each value.].
	^ array! !

!MostGeneralUnifier methodsFor: 'enumerating' stamp: 'fbs 1/12/2012 18:45'!
associationsDo: aBlock 
	"Evaluate aBlock for each of the receiver's elements (key/value 
	associations)."
	^ dict associationsDo: aBlock.! !

!MostGeneralUnifier methodsFor: 'accessing' stamp: 'fbs 1/12/2012 18:43'!
at: key
	^ dict at: key.! !

!MostGeneralUnifier methodsFor: 'accessing' stamp: 'fbs 1/12/2012 09:47'!
at: key put: value
	^ self class newFromPairs: (self asArray , {key. value}).! !

!MostGeneralUnifier methodsFor: 'testing' stamp: 'fbs 1/12/2012 18:50'!
includesKey: anObject
	^ dict includesKey: anObject.! !

!MostGeneralUnifier methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:12'!
isEmpty
	^ dict isEmpty.! !

!MostGeneralUnifier methodsFor: 'enumerating' stamp: 'fbs 2/10/2012 14:51'!
keys
	^ dict keys.! !

!MostGeneralUnifier methodsFor: 'enumerating' stamp: 'fbs 2/10/2012 15:03'!
keysMappingTo: anObject
	^ dict keys select: [:k | (self at: k) == anObject].! !

!MostGeneralUnifier methodsFor: 'running' stamp: 'fbs 6/14/2012 09:16'!
member: anObject of: someObjects
	^ self addAll: ([(HeadUnifier value: anObject) unifyWith: someObjects]
		on: UnificationFailure do:
			[[((HeadUnifier value: anObject) | #tail asVariable) unifyWith: someObjects]
				on: UnificationFailure do: [self]])! !

!MostGeneralUnifier methodsFor: 'printing' stamp: 'fbs 1/12/2012 19:09'!
printOn: aStream
	aStream
		nextPutAll: self class name.
	dict printElementsOn: aStream.! !

!MostGeneralUnifier methodsFor: 'accessing' stamp: 'fbs 1/12/2012 18:46'!
size
	^ dict size.! !

!MostGeneralUnifier methodsFor: 'enumerating' stamp: 'fbs 2/10/2012 14:51'!
values
	^ dict values.! !

!MostGeneralUnifier methodsFor: 'private' stamp: 'fbs 1/12/2012 18:43'!
wrappingDictionary: aDictionary
	dict := aDictionary.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/11/2012 20:33'!
failToUnifyWith: anObject
	UnificationFailure signalThis: self that: anObject.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/4/2012 11:32'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ anMGU.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:33'!
isMetaVariable
	^ false.! !

!Object methodsFor: '*Unification' stamp: 'fbs 7/23/2011 11:25'!
matchOneOf: aBlockAssociationCollection
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, report an error."

	^ self matchOneOf: aBlockAssociationCollection otherwise: [self caseError].! !

!Object methodsFor: '*Unification' stamp: 'fbs 8/8/2011 19:07'!
matchOneOf: someAssociations otherwise: aBlock
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, answer the result
	 of evaluating aBlock."
	^ (UnificationMatcher matcherUsing: someAssociations otherwise: aBlock) match: self.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/9/2012 18:54'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].

	^ anObject unificationClosureWithObject: self in: termRelation.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/9/2012 09:28'!
unificationClosureWithAlgebraic: anObject in: termRelation
	^ self unificationClosureWithSequence: anObject unapply in: termRelation.! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/11/2012 19:17'!
unificationClosureWithObject: anObject in: termRelation
	^ (self = anObject)
		ifTrue: [termRelation union: self with: anObject]
		ifFalse: ["s ~= t" self failToUnifyWith: anObject]! !

!Object methodsFor: '*Unification' stamp: 'fbs 1/11/2012 19:18'!
unificationClosureWithSequence: aSequence in: termRelation
	^ self failToUnifyWith: aSequence.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/26/2011 20:42'!
unifyWith: anObject
	^ Unifier unify: self with: anObject.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/8/2012 22:27'!
unifyWith: anObject in: aMostGeneralUnifier
	| newKnowledge |
	newKnowledge := self unifyWith: anObject.
	^ newKnowledge addAll: aMostGeneralUnifier.! !

Object subclass: #OrUnifier
	instanceVariableNames: 'left right'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!OrUnifier commentStamp: 'fbs 8/8/2011 22:31' prior: 0!
An OrUnifier is an object that allows an object to match one of two different objects:

OrUnifier
	left: (Node left: #x asVariable)
	right: (Node right: #x asVariable)
	
returns something that unifies with single-child Nodes.!

!OrUnifier class methodsFor: 'instance creation' stamp: 'fbs 8/8/2011 22:33'!
left: leftObject right: rightObject
	^ self new left: leftObject right: rightObject.! !

!OrUnifier methodsFor: 'private' stamp: 'fbs 8/8/2011 22:33'!
left: leftObject right: rightObject
	left := leftObject.
	right := rightObject.! !

!OrUnifier methodsFor: '*Unification' stamp: 'fbs 1/4/2012 20:31'!
unificationClosureWithAlgebraic: anObject in: termRelation
	[^ anObject unificationClosureWithAlgebraic: left in: termRelation]
		on: UnificationFailure do:
			[^ anObject unificationClosureWithAlgebraic: right in: termRelation]! !

!OrUnifier methodsFor: '*Unification' stamp: 'fbs 8/9/2011 07:36'!
unifyWith: anObject
	[^ left unifyWith: anObject]
		on: UnificationFailure do: [^ right unifyWith: anObject].! !

Object subclass: #SolutionFinder
	instanceVariableNames: 'acyclic visited partition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!SolutionFinder commentStamp: 'fbs 1/4/2012 10:29' prior: 0!
Given an equivalence relation (partition) on some set of nodes, each class potentially containing one or more MetaVariables, construct a MostGeneralUnifier.!

!SolutionFinder class methodsFor: 'instance creation' stamp: 'fbs 1/16/2012 15:50'!
findSolutionFor: aVariableAvoidingUnionFind
	"The original algorithm from Baader & Synder's _Unification Theory_ shows finding the solution starting only from the root of one of the two operands s and t to be unified. However, ListUnifier unifies on an artificial node - a node that we construct and does not itself appear in either s or t. Thus we merge the partial solutions found by starting from a representative of each of the equivalence classes in the partition."
	^ aVariableAvoidingUnionFind variableContainingClasses
		inject: MostGeneralUnifier new
		into: [:mgu :node |
			mgu addAll: (self new findSolutionFor: aVariableAvoidingUnionFind starting: node)]! !

!SolutionFinder methodsFor: 'private' stamp: 'fbs 1/12/2012 19:02'!
findSolution: anObject using: anMGU
	| newUnifier schemaTerm |
	schemaTerm := partition find: anObject.
	(acyclic includes: schemaTerm) ifTrue: [^ anMGU "s is not part of a cycle."].
	(visited includes: schemaTerm) ifTrue: [^ UnificationCycleFailure signalCycleOn: schemaTerm. "Cycle detected."].
	
	visited add: schemaTerm.
	newUnifier := anMGU addAll: (schemaTerm findSolutionWithUnifier: anMGU in: self).
	visited remove: schemaTerm.
	
	acyclic add: schemaTerm.
	^ (partition varsAt: schemaTerm)
		inject: newUnifier
		into: [:mgu :x | (x ~= schemaTerm)
			ifTrue: [mgu at: x put: schemaTerm]
			ifFalse: [mgu]].! !

!SolutionFinder methodsFor: 'unification' stamp: 'fbs 1/4/2012 11:45'!
findSolutionFor: aVariableAvoidingUnionFind starting: anObject
	partition := aVariableAvoidingUnionFind.
	acyclic := OrderedCollection new.
	visited := OrderedCollection new.
	^ self findSolution: anObject using: MostGeneralUnifier new.! !

Object subclass: #UnificationMatcher
	instanceVariableNames: 'cases otherwise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationMatcher commentStamp: 'fbs 7/28/2011 22:15' prior: 0!
I provide a unification-based case-like statement.

I use memoisation as an optimisation.!

!UnificationMatcher class methodsFor: 'instance creation' stamp: 'fbs 7/28/2011 22:29'!
matcherUsing: aBlockAssociationCollection
	^ self new matcherUsing: aBlockAssociationCollection.! !

!UnificationMatcher class methodsFor: 'instance creation' stamp: 'fbs 7/28/2011 22:14'!
matcherUsing: aBlockAssociationCollection otherwise: aBlock
	^ self new matcherUsing: aBlockAssociationCollection otherwise: aBlock.! !

!UnificationMatcher methodsFor: 'applying' stamp: 'fbs 6/14/2012 09:17'!
match: anObject
	"Iterate through the cases. As we evaluate the key thunks, replace them with the evaluated values. Much uglier than the naive implementation, but MUCH faster."
	| newCases result unmatch |
	result := unmatch := Object new.
	
	"As we try unify against the key thunk, build up a new Dictionary with the unifying key thunk memoised."
	newCases := OrderedCollection new.
	cases associationsDo:
		[:assoc | | mgu newKey |
			[newKey := assoc key isBlock ifTrue: [assoc key value] ifFalse: [assoc key].
			mgu := newKey unifyWith: anObject.
			result = unmatch ifTrue: [result := assoc value cull: mgu].
			newCases
				add: newKey;
				add: assoc value.] on: UnificationFailure do: ["nothing"]].

	cases := Dictionary newFromPairs: newCases.
	
	^ result == unmatch
		ifTrue: [otherwise value]
		ifFalse: [result].! !

!UnificationMatcher methodsFor: 'initialize-release' stamp: 'fbs 1/8/2012 22:05'!
matcherUsing: aBlockAssociationCollection
	self matcherUsing: aBlockAssociationCollection otherwise: [self caseError].! !

!UnificationMatcher methodsFor: 'initialize-release' stamp: 'fbs 8/8/2011 22:22'!
matcherUsing: aBlockAssociationCollection otherwise: aBlock
	cases := aBlockAssociationCollection asOrderedCollection.
	otherwise := aBlock.! !

Object subclass: #Unifier
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!Unifier commentStamp: 'fbs 5/2/2012 09:50' prior: 0!
I attempt to unify two structures, that is, find mappings between the MetaVariables (if any) within the two structures such that the structures represent the same structure.

I return a MostGeneralUnifier (which might be empty if the two structures are identical), or I raise a UnificationFailure.!

!Unifier class methodsFor: 'instance creation' stamp: 'fbs 1/11/2012 22:38'!
unify: left with: right
	| partition |
	partition := VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new.
	partition := (partition find: left)
		unificationClosureWith: (partition find: right) in: partition.
	^ SolutionFinder findSolutionFor: partition.! !

!Unifier class methodsFor: 'instance creation' stamp: 'fbs 2/9/2012 19:14'!
unify: left with: middle with: right
	| partition |
	partition := VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new.
	partition := (partition find: left)
		unificationClosureWith: (partition find: middle) in: partition.
	partition := (partition find: middle)
		unificationClosureWith: (partition find: right) in: partition.
		
	^ SolutionFinder findSolutionFor: partition.! !

Error subclass: #MostGeneralUnifierMergeFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!MostGeneralUnifierMergeFailure commentStamp: 'fbs 5/2/2012 10:05' prior: 0!
I represent the failure to merge two MostGeneralUnifiers with conflicting bindings - one says x -> 1 and the other says x -> 2, for instance.!

Error subclass: #UnificationFailure
	instanceVariableNames: 'this that'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationFailure commentStamp: 'fbs 1/11/2012 20:36' prior: 0!
I represent the inability to unify two patterns.

Note that in `a =? b`, my this instvar might reference a or b: unification makes no attempt to deterministically assign the left operand to either this or that. (We use double dispatch extensively, and sometimes double-double dispatch, so in general there's no way to correlate the operand handedness to the this or that instvars.)!

UnificationFailure subclass: #UnificationCycleFailure
	instanceVariableNames: 'cyclicTerm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationCycleFailure commentStamp: 'fbs 5/2/2012 09:48' prior: 0!
I represent the failure to unify a cyclic structure, such as x =? {x}.!

!UnificationCycleFailure class methodsFor: 'as yet unclassified' stamp: 'fbs 1/11/2012 20:09'!
signalCycleOn: aCyclicTerm
	^ self new
		cycle: aCyclicTerm;
		signal.! !

!UnificationCycleFailure methodsFor: 'private' stamp: 'fbs 1/11/2012 20:10'!
cycle: anObject
	cyclicTerm := anObject.! !

!UnificationCycleFailure methodsFor: 'accessing' stamp: 'fbs 1/11/2012 20:10'!
cyclicTerm
	^ cyclicTerm.! !

!UnificationFailure class methodsFor: 'instance creation' stamp: 'fbs 1/11/2012 20:32'!
signalThis: thisObject that: thatObject
	^ self new
		this: thisObject that: thatObject;
		signal.! !

!UnificationFailure methodsFor: 'accessing' stamp: 'fbs 1/11/2012 20:34'!
that
	^ that! !

!UnificationFailure methodsFor: 'accessing' stamp: 'fbs 1/11/2012 20:34'!
this
	^ this! !

!UnificationFailure methodsFor: 'private' stamp: 'fbs 1/11/2012 20:33'!
this: thisObject that: thatObject
	this := thisObject.
	that := thatObject.! !

PersistentUnionFind subclass: #VariableTrackingUnionFind
	instanceVariableNames: 'vars'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!VariableTrackingUnionFind commentStamp: 'fbs 1/4/2012 20:42' prior: 0!
A VariableAvoidingUnionFind is a persistent union-find structure that also tracks MetaVariables.

In particular, it maintains the invariants of a _schema function_ mapping equivalence classes to terms: a schema function S is a function such that
* S(C) elem C,
* S(C) is a MetaVariable only if every element in C is a MetaVariable.!

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/4/2012 13:31'!
at: anObject addVar: aMetaVariable
	| representative |
	representative := self find: anObject.
	^ vars at: representative put: ((vars at: representative) , {aMetaVariable})! !

!VariableTrackingUnionFind methodsFor: 'initialize-release' stamp: 'fbs 1/3/2012 16:10'!
initialize
	super initialize.
	vars := PersistentCollection initially: (DictionaryWithDefault wrapping: Dictionary new default: [:unused | Set new]).! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/3/2012 16:03'!
ranks: anArray parents: anotherArray vars: aThirdArray
	super ranks: anArray parents: anotherArray.
	vars := aThirdArray.! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/15/2012 21:53'!
storeSchemaTermFor: anObject as: anotherObject
	"The schema function is a function from terms to equivalence classes such that:
	* sigma(C) elem C,
	* sigma(C) is a variable only if every c in C is a variable."
	
	(anObject isMetaVariable and: [anotherObject isMetaVariable])
		ifTrue:
			[^ super storeSchemaTermFor: anObject as: anotherObject].

	anotherObject isMetaVariable
		ifTrue:
			[^ super storeSchemaTermFor: anotherObject as: anObject].

	^ super storeSchemaTermFor: anObject as: anotherObject.! !

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/15/2012 19:01'!
union: anObject with: anotherObject
	| cx cy rx ry |
	cx := self find: anObject.
	cy := self find: anotherObject.
	(cx = cy) ifTrue: [^ self].
	
	rx := ranks at: cx.
	ry := ranks at: cy.
	
	(rx > ry)
		ifTrue: [
			^ self class new ranks: (ranks at: cy put: ry + rx)
				parents: (self storeSchemaTermFor: cy as: cx)
				vars: (self updateVarsFor: cy and: cx)].
		
	^ self class new ranks: (ranks at: cx put: ry + rx)
		parents: (self storeSchemaTermFor: cx as: cy)
		vars: (self updateVarsFor: cx and: cy).! !

!VariableTrackingUnionFind methodsFor: 'private' stamp: 'fbs 1/15/2012 16:16'!
updateVarsFor: anObject and: anotherObject
	"The schema function is a function from terms to equivalence classes such that:
	* sigma(C) elem C,
	* sigma(C) is a variable only if every c in C is a variable."
	(anObject isMetaVariable and: [anotherObject isMetaVariable])
		ifTrue:
			[^ self at: anotherObject addVar: anObject].

	anotherObject isMetaVariable
		ifTrue:
			[^ self at: anObject addVar: anotherObject].

	anObject isMetaVariable
		ifTrue:
			[^ self at: anotherObject addVar: anObject].
		
	^ vars.! !

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/15/2012 20:12'!
variableContainingClasses
	"Return the representatives of those classes containing variables."
	^ vars withValue: [:dict |
		(dict reject: #isEmpty) keys].! !

!VariableTrackingUnionFind methodsFor: 'accessing' stamp: 'fbs 1/3/2012 16:01'!
varsAt: anObject
	^ vars at: anObject! !

!VariableTrackingUnionFind methodsFor: 'enumerating' stamp: 'fbs 1/15/2012 15:34'!
varsInject: initialValue into: binaryBlock
	^ (vars inject: Set new into: [:allVars :theseVars | allVars , theseVars])
		inject: initialValue into: binaryBlock.! !

!Symbol methodsFor: '*Unification' stamp: 'fbs 7/22/2011 13:18'!
asVariable
	^ MetaVariable named: self.! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/12/2012 09:40'!
findSolutionWithUnifier: anMGU in: solutionFinder
	^ self
		inject: anMGU
		into: [:mgu :each | solutionFinder findSolution: each using: mgu].! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/3/2012 15:30'!
unificationClosureWith: anObject in: termRelation
	anObject isMetaVariable
		ifTrue: [^ termRelation union: self with: anObject].
		
	^ anObject unificationClosureWithSequence: self in: termRelation.! !

!SequenceableCollection methodsFor: '*Unification' stamp: 'fbs 1/11/2012 19:18'!
unificationClosureWithSequence: aCollection in: termRelation
	| rel |
	(self size = aCollection size) ifFalse: [^ self failToUnifyWith: aCollection].
	rel := termRelation.
	self with: aCollection do:
		[:si :ti | rel := (rel find: si) unificationClosureWith: (rel find: ti) in: rel].
	^ rel.! !

TestCase subclass: #HeadUnifierTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests-Destructuring'!

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:14'!
testCanUnifyAgainstFirstNWhenLeftOperand
	| mgu |
	mgu := (HeadUnifier values: {#x asVariable. #y asVariable}) unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: 2 equals: (mgu at: #y asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:14'!
testCanUnifyAgainstFirstNWhenRightOperand
	| mgu |
	mgu := #(1 2 3) unifyWith: (HeadUnifier values: {#x asVariable. #y asVariable}).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: 2 equals: (mgu at: #y asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:14'!
testCanUnifyAgainstaVariable
	| mgu head |
	mgu := #x asVariable unifyWith: (HeadUnifier named: #y).
	head := mgu at: #x asVariable.
	self assert: #y asVariable equals: head value.! !

!HeadUnifierTest methodsFor: 'testing - error reporting' stamp: 'fbs 6/14/2012 09:14'!
testErrorMentionsUnunifiableNodes
	[1 unifyWith: (HeadUnifier named: #y)] on: UnificationFailure do:
		[:e |
		self assert: (e this isKindOf: HeadUnifier).
		self assert: 1 equals: e that].
	
	[(HeadUnifier named: #y) unifyWith: 1] on: UnificationFailure do:
		[:e |
		self assert: (e this isKindOf: HeadUnifier).
		self assert: 1 equals: e that].! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:14'!
testHeadCantUnifyWithNormalObjects
	self should: [1 unifyWith: (HeadUnifier named: #x)] raise: UnificationFailure.
	self should: [$a unifyWith: (HeadUnifier named: #x)] raise: UnificationFailure.
	self should: [Object new unifyWith: (HeadUnifier named: #x)] raise: UnificationFailure.
	self should: [(HeadUnifier named: #x) unifyWith: 1] raise: UnificationFailure.
	self should: [(HeadUnifier named: #x) unifyWith: $a] raise: UnificationFailure.
	self should: [(HeadUnifier named: #x) unifyWith: Object new] raise: UnificationFailure.! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:14'!
testPipeConstructsaListUnifier
	| list mgu |
	list := (HeadUnifier named: #x) | #y asVariable.
	self assert: ListUnifier equals: list class.
	mgu := list unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testUnifyingWithCollectionUnifiesOnFirstElement
	| mgu |
	mgu := (HeadUnifier named: #x) unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testUnifyingWithCollectionUnifiesOnFirstElementWhenCollection
	| mgu |
	mgu := (HeadUnifier value: {#x asVariable}) unifyWith: #((1) 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testUnifyingWithCollectionUnifiesOnFirstElementWhenRightOperand
	| mgu |
	mgu := #(1 2 3) unifyWith: (HeadUnifier named: #x).
	self assert: 1 equals: (mgu at: #x asVariable).! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testUnifyingWithEmptyCollectionFails
	self should: [(HeadUnifier named: #x) unifyWith: #()] raise: UnificationFailure.! !

!HeadUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testUnifyingWithEmptyCollectionFailsWhenRightOperand
	self should: [#() unifyWith: (HeadUnifier named: #x)] raise: UnificationFailure.! !

TestCase subclass: #ListUnifierTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests-Destructuring'!

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testCanAppearOnRightHandSide
	| mgu |
	mgu := #(1 2 3) unifyWith: (ListUnifier headNamed: #x tailNamed: #y).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testCanUnifyAgainstaVariable
	| mgu list |
	mgu := #z asVariable unifyWith: (ListUnifier headNamed: #x tailNamed: #y).
	list := mgu at: #z asVariable.
	self assert: #x asVariable equals: list head.
	self assert: #y asVariable equals: list tail.! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:15'!
testCantUnifyWithNormalObjects
	self should: [1 unifyWith: (ListUnifier headNamed: #x tailNamed: #y)] raise: UnificationFailure.
	self should: [$a unifyWith: (ListUnifier headNamed: #x tailNamed: #y)] raise: UnificationFailure.
	self should: [Object new unifyWith: (ListUnifier headNamed: #x tailNamed: #y)] raise: UnificationFailure.
	self should: [(ListUnifier headNamed: #x tailNamed: #y) unifyWith: 1] raise: UnificationFailure.
	self should: [(ListUnifier headNamed: #x tailNamed: #y) unifyWith: $a] raise: UnificationFailure.
	self should: [(ListUnifier headNamed: #x tailNamed: #y) unifyWith: Object new] raise: UnificationFailure.! !

!ListUnifierTest methodsFor: 'testing - error reporting' stamp: 'fbs 6/14/2012 09:15'!
testErrorMentionsUnunifiableNodes
	[1 unifyWith: (ListUnifier headNamed: #x tailNamed: #y)] on: UnificationFailure do:
		[:e |
		self assert: (e this isKindOf: ListUnifier).
		self assert: 1 equals: e that].! !

!ListUnifierTest methodsFor: 'testing - instance creation' stamp: 'fbs 6/14/2012 09:15'!
testHeadNamedTailNamedReturnsListUnifier
	| l mgu |
	l := ListUnifier headNamed: #x tailNamed: #y.
	mgu := l unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - instance creation' stamp: 'fbs 6/14/2012 09:15'!
testHeadNamedTailReturnsListUnifier
	| l mgu |
	l := ListUnifier headNamed: #x tail: #(2 3).
	mgu := l unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).! !

!ListUnifierTest methodsFor: 'testing - instance creation' stamp: 'fbs 6/14/2012 09:15'!
testHeadTailNamedReturnsListUnifier
	| l mgu |
	l := ListUnifier head: 1 tailNamed: #x.
	mgu := l unifyWith: #(1 2 3).
	self assert: #(2 3) equals: (mgu at: #x asVariable).! !

!ListUnifierTest methodsFor: 'testing - instance creation' stamp: 'fbs 6/14/2012 09:16'!
testHeadTailReturnsListUnifier
	| l mgu |
	l := ListUnifier head: #x asVariable tail: #y asVariable.
	mgu := l unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithCollectionDestructuresIntoHeadAndTail
	| mgu |
	mgu := (ListUnifier headNamed: #x tailNamed: #y) unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithCollectionPermitsMultiElementHead
	| mgu |
	"[X, Y | T] = [1, 2, 3] in Prolog"
	mgu := (ListUnifier head: {#x asVariable. #y asVariable} tail: #t asVariable) unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: 2 equals: (mgu at: #y asVariable).
	self assert: #(3) equals: (mgu at: #t asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithCollectionPermitsMultiElementHeadEvenWithEmptyTail
	| mgu |
	"[X, Y | T] = [1, 2, 3] in Prolog"
	mgu := (ListUnifier head: {#x asVariable. #y asVariable} tail: #t asVariable)
		unifyWith: (OrderedCollection with: 1 with: 2).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: 2 equals: (mgu at: #y asVariable).
	self assert: OrderedCollection new equals: (mgu at: #t asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithSingleElementCollectionMakesTailEmpty
	| mgu |
	mgu := (ListUnifier headNamed: #x tailNamed: #y) unifyWith: #(1).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #() equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithSingleElementCollectionRespectsCollectionType
	| mgu |
	mgu := (ListUnifier headNamed: #x tailNamed: #y) unifyWith: (OrderedCollection with: 1).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: OrderedCollection new equals: (mgu at: #y asVariable).! !

!ListUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testUnifyWithSpecificTail
	| mgu |
	mgu := (ListUnifier headNamed: #x tail: {#y asVariable. #z asVariable}) unifyWith: #(1 2 3).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: 2 equals: (mgu at: #y asVariable).
	self assert: 3 equals: (mgu at: #z asVariable).! !

TestCase subclass: #MatchingTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!MatchingTest methodsFor: 'helpers' stamp: 'fbs 1/9/2012 09:38'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1}, expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!MatchingTest methodsFor: 'testing - helpers' stamp: 'fbs 1/9/2012 09:38'!
testAssertEqualsDescriptionGivesMeaningfulErrorMessage
	[self assert: 1 equals: 2 description: 'Fail!!'] on: TestFailure do:
		[ :failure |
		self assert: 'Fail!!, expected 1, got 2' equals: failure messageText].! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 14:18'!
testEmptyMguPassedIntoValueBlockMatchWins
	| result substitution |
	result := 1 matchOneOf: {[1] -> [:mgu | substitution := mgu. 'foo']} otherwise: ['bar'].
	self assert: result = 'foo'.
	self assert: substitution isEmpty.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:34'!
testFirstMatchWins
	| result |
	result := 1
		matchOneOf:
			{[1] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'first match' description: 'Unexpected match invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:35'!
testMatchPassesUnifierIntoBlock
	| result mgu |
	result := {1}
		matchOneOf:
			{[{#x asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) = 1 description: 'Unexpected MGU'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:36'!
testMatchingTriesAllBlocks
	| result |
	result := 1
		matchOneOf:
			{[0] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'second match' description: 'Unexpected result'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:32'!
testMatchingUnifiesBidirectionally
	| result mgu |
	result := {#x asVariable. 1}
		matchOneOf:
			{[{2. #y asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) equals: 2 description: 'Unexpected MGU (x)'.
	self assert: (mgu at: #y asVariable) equals: 1 description: 'Unexpected MGU (y)'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := 1
		matchOneOf:
			{['not a 1'] -> ['matched something']}
		otherwise: ['otherwise block'].
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesRaisesError
	self should: [1
		matchOneOf:
			{['not a 1'] -> ['matched something']}]
		raise: Error.! !

TestCase subclass: #MetaVariableTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:34'!
testMetaVariableEqualToSameNamedMetaVariable
	self assert: #x asVariable equals: #x asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableHashEqualToSameNamedMetaVariableHash
	self assert: #x asVariable hash equals: #x asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:36'!
testMetaVariableHashNotEqualToDifferentlyNamedMetaVariableHash
	self deny: #x asVariable hash = #X asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToDifferentlyNamedMetaVariable
	self deny: #x asVariable = #X asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToNil
	self deny: #x asVariable = nil! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 11:03'!
testUnapplyShouldDestructureCorrectly
	self assert: (MetaVariable named: #a) unapply = {#Variable. #a.}! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:08'!
testVariableHasNoChildren
	self deny: (MetaVariable named: #foo) hasChildren.

	self assert: (MetaVariable named: #foo) children isEmpty.! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testVariableIsVariable
	self assert: (MetaVariable named: #foo) isMetaVariable.! !

TestCase subclass: #MostGeneralUnifierTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:37'!
testAddAllAddsOldKeys
	| old new |
	old := MostGeneralUnifier new.
	new := old addAll: (MostGeneralUnifier newFromPairs: #(1 2)).
	self assert: (new includesKey: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:37'!
testAddAllLeavesOriginalInstanceUntouched
	| old new |
	old := MostGeneralUnifier new.
	new := old addAll: (MostGeneralUnifier newFromPairs: #(1 2)).
	self deny: (old includesKey: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 2/10/2012 14:41'!
testAddAllMergesTwoStepLookups
	"x -> y, y -> 1 => x -> 1, y -> 1"
	| old new |
	old := MostGeneralUnifier newFromPairs: {#x asVariable. #y asVariable}.
	new := old addAll: (MostGeneralUnifier newFromPairs: {#y asVariable. 1}).
	self assert: 1 equals: (new at: #x asVariable).! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:45'!
testAddAllReturnsNewMGU
	| old new |
	old := MostGeneralUnifier new.
	new := old addAll: (MostGeneralUnifier newFromPairs: #(1 2)).
	self deny: old = new.! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 5/2/2012 10:05'!
testAddAllWithConflictingBindingBlowsUp
	| old new |
	old := MostGeneralUnifier newFromPairs: #(1 2).
	self
		should: [new := old addAll: (MostGeneralUnifier newFromPairs: #(1 3))]
		raise: MostGeneralUnifierMergeFailure
		whoseDescriptionIncludes: '1'
		description: 'Unexpected exception, or poorly described exception'.! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:46'!
testAtPutAddsKeyValue
	| old new |
	old := MostGeneralUnifier new.
	new := old at: 1 put: 2.
	self assert: (new includesKey: 1).
	self assert: 2 equals: (new at: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:46'!
testAtPutLeavesOriginalInstanceUntouched
	| old new |
	old := MostGeneralUnifier new.
	new := old at: 1 put: 2.
	self deny: (old includesKey: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - adding' stamp: 'fbs 1/12/2012 09:45'!
testAtPutReturnsNewMGU
	| old new |
	old := MostGeneralUnifier new.
	new := old at: 1 put: 2.
	self deny: old = new.! !

!MostGeneralUnifierTest methodsFor: 'testing - accessing' stamp: 'fbs 1/12/2012 18:52'!
testAtReturnsValueForSomeKey
	self assert: 2 equals: ((MostGeneralUnifier new at: 1 put: 2) at: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - accessing' stamp: 'fbs 1/12/2012 18:53'!
testAtThrowsExceptionForMissingKey
	self should: [MostGeneralUnifier new at: 1] raise: KeyNotFound.! !

!MostGeneralUnifierTest methodsFor: 'testing - instance creation' stamp: 'fbs 5/2/2012 16:45'!
testEmptyMGUaConstantValue
	self assert: MostGeneralUnifier empty == MostGeneralUnifier empty.! !

!MostGeneralUnifierTest methodsFor: 'testing - testing' stamp: 'fbs 1/12/2012 18:51'!
testIncludesKeyReturnsFalseWhenNoKeyPresent
	self deny: (MostGeneralUnifier new includesKey: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - testing' stamp: 'fbs 1/12/2012 18:52'!
testIncludesKeyReturnsTrueWhenKeyPresent
	self assert: ((MostGeneralUnifier new at: 1 put: 2) includesKey: 1).! !

!MostGeneralUnifierTest methodsFor: 'testing - testing' stamp: 'fbs 1/12/2012 19:11'!
testIsEmptyReturnsFalseWhenNotEmpty
	self deny: (MostGeneralUnifier new at: 1 put: 2) isEmpty.! !

!MostGeneralUnifierTest methodsFor: 'testing - testing' stamp: 'fbs 1/12/2012 19:11'!
testIsEmptyReturnsTrueWhenEmpty
	self assert: MostGeneralUnifier new isEmpty.! !

!MostGeneralUnifierTest methodsFor: 'testing - printing' stamp: 'fbs 1/12/2012 20:27'!
testPrintOnEmptyMGUPrintsNicely
	| mgu |
	mgu := MostGeneralUnifier new.
	self assert: 'MostGeneralUnifier()' equals: mgu printString.! !

!MostGeneralUnifierTest methodsFor: 'testing - printing' stamp: 'fbs 1/12/2012 20:27'!
testPrintOnPrintsAlmostLikeDictionary
	| mgu |
	mgu := MostGeneralUnifier newFrom: {1 -> 2}.
	self assert: 'MostGeneralUnifier(1->2 )' equals: mgu printString.! !

!MostGeneralUnifierTest methodsFor: 'testing - accessing' stamp: 'fbs 1/12/2012 18:55'!
testSizeReturnsNumberOfKeys
	#(1 2 3 4)
		inject: MostGeneralUnifier new
		into: [:mgu :i | | m |
			m := mgu at: i put: i.
			self assert: i equals: m size.
			m].! !

TestCase subclass: #OrUnifierTest
	instanceVariableNames: 'matcher'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!OrUnifierTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:45'!
setUp
	matcher := OrUnifier
		left: (TreeNode left: #x asVariable)
		right: (TreeNode right: #x asVariable).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 6/14/2012 09:16'!
testFirstOptionUnifies
	| mgu |
	mgu := matcher unifyWith: (TreeNode left: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 6/14/2012 09:16'!
testNoUnificationPossibleStillSignalsError
	self should: [matcher unifyWith: Tree empty] raise: UnificationFailure! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 6/14/2012 09:16'!
testSecondOptionUnifies
	| mgu |
	mgu := matcher unifyWith: (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 6/14/2012 09:16'!
testUnifiesAsRightHandOperandAgainstFirstOption
	| mgu |
	mgu := (TreeNode left: (Leaf value: 1)) unifyWith: matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 6/14/2012 09:16'!
testUnifiesAsRightHandOperandAgainstSecondOption
	| mgu |
	mgu := (TreeNode right: (Leaf value: 1)) unifyWith: matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

TestCase subclass: #SolutionFinderTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!SolutionFinderTest commentStamp: 'fbs 1/4/2012 10:39' prior: 0!
Most of SolutionFinder's tests are found in UnificationTest.!

!SolutionFinderTest methodsFor: 'fixtures' stamp: 'fbs 1/5/2012 09:28'!
makePartition
	| partition |
	partition := VariableTrackingUnionFind usingArrayType: PersistentCollection partitioning: Dictionary new.
	partition := (partition find: 1) unificationClosureWith: (partition find: #x asVariable) in: partition.
	^ partition.! !

!SolutionFinderTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 22:43'!
testCanCreateFinder
	| partition |
	partition := self makePartition.
	SolutionFinder findSolutionFor: partition.! !

TestCase subclass: #TailUnifierTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests-Destructuring'!

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:16'!
testCanUnifyAgainstaVariable
	| mgu tail |
	mgu := #x asVariable unifyWith: (TailUnifier named: #y).
	tail := mgu at: #x asVariable.
	self assert: #y asVariable equals: tail value.! !

!TailUnifierTest methodsFor: 'testing - error reporting' stamp: 'fbs 6/14/2012 09:17'!
testErrorMentionsUnunifiableNodes
	[1 unifyWith: (TailUnifier named: #y)] on: UnificationFailure do:
		[:e |
		self assert: (e this isKindOf: TailUnifier).
		self assert: 1 equals: e that].! !

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:17'!
testTailCantUnifyWithNormalObjects
	self should: [1 unifyWith: (TailUnifier named: #x)] raise: UnificationFailure.
	self should: [$a unifyWith: (TailUnifier named: #x)] raise: UnificationFailure.
	self should: [Object new unifyWith: (TailUnifier named: #x)] raise: UnificationFailure.
	self should: [(TailUnifier named: #x) unifyWith: 1] raise: UnificationFailure.
	self should: [(TailUnifier named: #x) unifyWith: $a] raise: UnificationFailure.
	self should: [(TailUnifier named: #x) unifyWith: Object new] raise: UnificationFailure.! !

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:17'!
testUnifyingWithCollectionUnifiesOnAllButFirstElements
	| mgu |
	mgu := (TailUnifier named: #x) unifyWith: #(1 2 3).
	self assert: #(2 3) equals: (mgu at: #x asVariable).! !

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:17'!
testUnifyingWithCollectionUnifiesOnAllButFirstElementsWhenRightOperand
	| mgu |
	mgu := #(1 2 3) unifyWith: (TailUnifier named: #x).
	self assert: #(2 3) equals: (mgu at: #x asVariable).! !

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:17'!
testUnifyingWithEmptyCollectionFails
	self should: [(TailUnifier named: #x) unifyWith: #()] raise: UnificationFailure.! !

!TailUnifierTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:17'!
testUnifyingWithEmptyCollectionFailsWhenRightOperand
	self should: [#() unifyWith: (TailUnifier named: #x)] raise: UnificationFailure.! !

TestCase subclass: #UnificationCycleFailureTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationCycleFailureTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 20:08'!
testSignalCycleOnRaisesException
	self should: [UnificationCycleFailure signalCycleOn: 1] raise: UnificationCycleFailure.! !

!UnificationCycleFailureTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 20:08'!
testSignalCycleOnStoresCyclicTerm
	[UnificationCycleFailure signalCycleOn: 1] on: UnificationCycleFailure do:
		[:e | self assert: 1 equals: e cyclicTerm].! !

TestCase subclass: #UnificationFailureTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationFailureTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 20:38'!
testFailToUnifyWithRaisesException
	[1 failToUnifyWith: 2]
		on: UnificationFailure do: [ :e |
			self assert: 1 equals: e this.
			self assert: 2 equals: e that].! !

!UnificationFailureTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 20:39'!
testSignalContainsUnunifiableObjects
	[UnificationFailure signalThis: 1 that: 2]
		on: UnificationFailure do: [ :e |
			self assert: 1 equals: e this.
			self assert: 2 equals: e that].! !

!UnificationFailureTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/11/2012 20:33'!
testSignalLeftRightRaisesException
	self should: [UnificationFailure signalThis: 1 that: 2] raise: UnificationFailure.! !

TestCase subclass: #UnificationMatcherTest
	instanceVariableNames: 'matcher substitution'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationMatcherTest methodsFor: 'helpers' stamp: 'fbs 7/28/2011 22:17'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1} , expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!UnificationMatcherTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:44'!
setUp
	matcher := UnificationMatcher matcherUsing:
		{[Leaf value: 1] -> ['first leaf'].
		[Leaf value: 1] -> ['second leaf'].
		[Tree empty] -> ['tree'].
		[TreeNode left: #x asVariable] -> [:mgu | substitution := mgu. 'mgu-using case']}
		otherwise: ['otherwise block'].! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 19:09'!
testFirstMatchWins
	| result |
	result := matcher match: (Leaf value: 1).
	self assert: result = 'first leaf' description: 'Unexpected match invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 10/7/2011 07:44'!
testMatchPassesUnifierIntoBlock
	| result |
	result := matcher match: (TreeNode left: #y asVariable).
	
	self assert: result = 'mgu-using case' description: 'No match found'.
	self assert: substitution notNil description: 'MGU not passed into block'.
	self assert: (substitution at: #y asVariable) = (#x asVariable) description: 'Unexpected MGU'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testMatchingTriesAllBlocks
	| result |
	result := matcher match: Tree empty.
	self assert: result = 'tree' description: 'Unexpected result'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := matcher match: (Leaf value: 2).
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 22:28'!
testNoMatchEvaluatesRaisesError
	| m |
	m := UnificationMatcher matcherUsing:
			{[Tree empty] -> ['matched something']}.
			
	self should: [m matchAgainst: (Leaf value: 1)] raise: Error.! !

TestCase subclass: #UnificationOfAlgebraicDataTypeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - extensions' stamp: 'fbs 6/14/2012 09:17'!
testAlgebraicDataTypeOrReturnsOrMatcher
	| mgu |
	mgu := ((TreeNode left: #x asVariable) or: (TreeNode right: #x asVariable)) unifyWith: (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - unification' stamp: 'fbs 1/5/2012 09:24'!
testDifferentArityTermsShouldNotUnify
	"f(x) ~= g(x,y)"
	|f g |
	f := VariadicTerm with: #(#x).
	g := VariadicTerm with: #(#x #y).
	self should: [f unifyWith: g] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - error reporting' stamp: 'fbs 6/14/2012 09:17'!
testErrorMentionsUnunifiableNodes
	[1 unifyWith: (Leaf value: 1)] on: UnificationFailure do:
		[:e |
		self assert: (Leaf value: 1) equals: e this.
		self assert: 1 equals: e that].
	
	[(Leaf value: 1) unifyWith: 1] on: UnificationFailure do:
		[:e |
		self assert: 1 equals: e this.
		self assert: (Leaf value: 1) equals: e that].! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 6/14/2012 09:17'!
testUnappliedAlgebraicDataTypesCantUnifyWithAlgebraicDataTypes
	self shouldnt: [ (Leaf value: 1) unifyWith: #(Leaf 1) ] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 6/14/2012 09:17'!
testUnificationWithCollectionsWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} unifyWith: {VariadicTerm with: {#a asVariable}. #y asVariable.}.
	self assert: (VariadicTerm with: {#a asVariable}) equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationOfCollectionsTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationOfCollectionsTest methodsFor: 'testing - error reporting' stamp: 'fbs 6/14/2012 09:18'!
testErrorMentionsUnunifiableNodes
	[{1} unifyWith: 1] on: UnificationFailure do:
		[:e |
		self assert: 1 equals: e this.
		self assert: {1} equals: e that].
	
	[1 unifyWith: {1}] on: UnificationFailure do:
		[:e |
		self assert: {1} equals: e this.
		self assert: 1 equals: e that].! !

!UnificationOfCollectionsTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:18'!
testNormalObjectsDontUnifyWithCollections
	self should: [1 unifyWith: #()] raise: UnificationFailure.
	self should: [#() unifyWith: 1] raise: UnificationFailure.! !

!UnificationOfCollectionsTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:18'!
testUnificationWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} unifyWith: {2. #y asVariable}.
	self assert: 2 equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:18'!
testUnificationWorksLeftward
	| mgu |
	mgu := {#x asVariable. #y asVariable.} unifyWith: #(1 (2 3)).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing - unification' stamp: 'fbs 6/14/2012 09:18'!
testUnificationWorksRightward
	| mgu |
	mgu := #(1 (2 3)) unifyWith: {#x asVariable. #y asVariable.}.
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!UnificationTest methodsFor: 'testing' stamp: 'fbs 6/14/2012 09:18'!
testAnythingCanUnifyAgainstMetaVariables
	self shouldnt: ['a' unifyWith: #x asVariable] raise: UnificationFailure.
	self shouldnt: [$a unifyWith: #x asVariable] raise: UnificationFailure.
	self shouldnt: [1 unifyWith: #x asVariable] raise: UnificationFailure.
	self shouldnt: [#() unifyWith: #x asVariable] raise: UnificationFailure.
	self shouldnt: [Object new unifyWith: #x asVariable] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/11/2012 22:09'!
testCyclicStructuresCantUnify
	| y |
	y := MetaVariable named: #y.
	self should: [y unifyWith: {y}] raise: UnificationFailure.
	[y unifyWith: {y}]
		on: UnificationCycleFailure
		do: [:e | self assert: {y} equals: e cyclicTerm].! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:16'!
testImpossibleUnificationShouldNotUnify
	self should: [1 unifyWith: 2] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 6/14/2012 09:18'!
testMetaVariablesCanUnifyAgainstAnything
	self shouldnt: [#x asVariable unifyWith: 'a'] raise: UnificationFailure.
	self shouldnt: [#x asVariable unifyWith: $a] raise: UnificationFailure.
	self shouldnt: [#x asVariable unifyWith: 1] raise: UnificationFailure.
	self shouldnt: [#x asVariable unifyWith: #()] raise: UnificationFailure.
	self shouldnt: [#x asVariable unifyWith: Object new] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:08'!
testSelfUnificationShouldUnify
	| x |
	x := MetaVariable named: #x.
	"Tautologies"
	self assert: (1 unifyWith: 1) isEmpty.
	self assert: (x unifyWith: x) isEmpty.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:18'!
testTrivialBidirectionalUnificationShouldUnify
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := x unifyWith: y.
	self assert: matches = (MostGeneralUnifier newFrom: {x -> y}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:12'!
testTrivialUnificationsShouldUnify
	| x |
	x := MetaVariable named: #x.
	self assert: (x unifyWith: 2) = (MostGeneralUnifier newFrom: {x -> 2}).
	self assert: (2 unifyWith: x) = (MostGeneralUnifier newFrom: {x -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:18'!
testTwoVariableBidirectionalUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. 2} unifyWith: {1. right}.
	self assert: matches = (MostGeneralUnifier newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:17'!
testTwoVariableUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. right} unifyWith: {1. 2}.
	self assert: matches = (MostGeneralUnifier newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/12/2012 19:17'!
testUnificationWorksOnDeepStructures
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := { {{x}} } unifyWith: {y}.
	self assert: matches = (MostGeneralUnifier newFrom: { y -> {{x}} }).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/15/2012 22:23'!
testVariableUnifyingAgainstTwoVariables
	| matches x y z |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	z := MetaVariable named: #z.
	matches := {x. x.} unifyWith: {y. z.}.
	self assert: matches = (MostGeneralUnifier newFrom: {y -> x. z -> x.}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 6/14/2012 09:18'!
testVariablesInsideCollectionsUnify
	| matches |
	matches := {1. 2. #x asVariable. 4} unifyWith: #(1 2 3 4).
	self assert: matches equals: (MostGeneralUnifier newFromPairs: {#x asVariable. 3.}).
	
	matches := #(1 2 3 4) unifyWith: {1. 2. #x asVariable. 4}.
	self assert: matches equals: (MostGeneralUnifier newFromPairs: {#x asVariable. 3.}).! !

TestCase subclass: #VariableTrackingUnionFindTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification-Tests'!

!VariableTrackingUnionFindTest methodsFor: 'fixtures' stamp: 'fbs 1/4/2012 20:26'!
makePartition
	^ VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new! !

!VariableTrackingUnionFindTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 16:25'!
testAddingNonVariableToVariableClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: 1.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 16:25'!
testAddingVariableToClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: 1 with: #x asVariable.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 16:26'!
testAddingVariableToVariableClassMakesOriginalVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: #y asVariable.
	self assert: #y asVariable equals: (v find: #y asVariable).
	self assert: #y asVariable equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'testing' stamp: 'fbs 1/15/2012 15:56'!
testUnifyingWithVariableLeavesOriginalInstanceUnchanged
	| p1 p2 |
	p1 := self makePartition.
	p2 := p1 union: 1 with: #x asVariable.
	self assert: #() equals: p1 variableContainingClasses asArray.! !

!VariableTrackingUnionFindTest methodsFor: 'testing - accessing' stamp: 'fbs 1/15/2012 15:46'!
testVariableContainingClassesReturnsCollectionOnNonVariablePartition
	| partition |
	partition := self makePartition union: #x asVariable with: 1.
	partition := partition union: #y asVariable with: 2.
	partition := partition union: 3 with: 4.
	self assert: #(1 2) sort equals: partition variableContainingClasses asArray sort.! !

!VariableTrackingUnionFindTest methodsFor: 'testing - accessing' stamp: 'fbs 1/15/2012 15:46'!
testVariableContainingClassesReturnsEmptyCollectionOnNonVariablePartition
	self assert: #() equals: (self makePartition union: 1 with: 2) variableContainingClasses asArray.! !

!VariableTrackingUnionFindTest methodsFor: 'testing - enumerating' stamp: 'fbs 1/14/2012 10:45'!
testVarsInjectIntoIteratesOverVariables
	| mgu vars |
	mgu := VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new.
	mgu := mgu union: #x asVariable with: 1.
	mgu := mgu union: #z asVariable with: 1.
	mgu := mgu union: #y asVariable with: 2.
	vars := mgu varsInject: Set new into: [:acc :var | acc , {var}].
	self assert: {#x asVariable. #y asVariable. #z asVariable} asSet equals: vars.! !

!VariableTrackingUnionFindTest methodsFor: 'testing - enumerating' stamp: 'fbs 1/13/2012 18:51'!
testVarsInjectIntoOverEmptyVariableCollectionDoesNothing
	| mgu vars |
	mgu := VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new.
	vars := mgu varsInject: #() into: [:acc :var | acc , var].
	self assert: #() equals: vars.! !
MostGeneralUnifier initialize!
